import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { Inject } from '@nestjs/common';
import {
  AggregateID,
  ConflictException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import { USERNAME_REPOSITORY } from '@modules/authentication/authentication.di-tokens';
import { UsernameAlreadyExistsException } from '@modules/authentication/core/domain/authentication.errors';
import { UpdateUsernameCommand } from './update-username.command';
import { UsernameRepositoryPort } from '../../ports/username.repository.port';
import { UsernameEntity } from '@modules/authentication/core/domain/username.entity';

@CommandHandler(UpdateUsernameCommand)
export class UpdateUsernameService implements ICommandHandler {
  constructor(
    @Inject(USERNAME_REPOSITORY)
    private readonly usernameRepository: UsernameRepositoryPort,
  ) {}

  async execute(command: UpdateUsernameCommand): Promise<AggregateID> {
    try {
      const username: UsernameEntity = await this.usernameRepository.findByType(
        command.userId,
        command.username.type,
      );
      const oldName: string = username.id;
      username.update({
        name: command.username.name,
      });
      await this.usernameRepository.updateUsername(oldName, username);
      return username.getProps().name;
    } catch (error: any) {
      if (error instanceof ConflictException) {
        throw new UsernameAlreadyExistsException(error);
      }
      if (
        error instanceof UniqueConstraintException &&
        error.message.includes('username')
      ) {
        throw new UsernameAlreadyExistsException(error);
      }
      throw error;
    }
  }
}
