import {
  AggregateID,
  ConflictException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import {
  AUTHENTICATION_REPOSITORY,
  USERNAME_REPOSITORY,
} from '@modules/authentication/authentication.di-tokens';
import { AddUsernameCommand } from '@modules/authentication/core/application/commands/add-username/add-username.command';
import { AddUsernameService } from '@modules/authentication/core/application/commands/add-username/add-username.service';
import { UsernameAlreadyExistsException } from '@modules/authentication/core/domain/authentication.errors';
import { UsernameEntity } from '@modules/authentication/core/domain/username.entity';
import { Type } from '@modules/authentication/core/domain/username.types';
import { AddUsernameRequestDto } from '@modules/authentication/interface/grpc-controllers/dtos/add-username.request.dto';
import { Test, TestingModule } from '@nestjs/testing';

const addUsernameRequest: AddUsernameRequestDto = {
  userId: '165192d4-398a-4469-a16b-98c02cc6f531',
  name: 'john.doe@email.com',
  type: Type.EMAIL,
};

const mockAuthenticationRepository = {
  findOneById: jest.fn().mockImplementationOnce(() => ({})),
};

const mockUsernameRepository = {
  insert: jest
    .fn()
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => {
      throw new ConflictException('already exists');
    })
    .mockImplementationOnce(() => {
      throw new UniqueConstraintException('username');
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Add Username Service', () => {
  let addUsernameService: AddUsernameService;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: AUTHENTICATION_REPOSITORY,
          useValue: mockAuthenticationRepository,
        },
        {
          provide: USERNAME_REPOSITORY,
          useValue: mockUsernameRepository,
        },
        AddUsernameService,
      ],
    }).compile();

    addUsernameService = module.get<AddUsernameService>(AddUsernameService);
  });

  it('should be defined', () => {
    expect(addUsernameService).toBeDefined();
  });

  describe('execution', () => {
    const addUsernameCommand = new AddUsernameCommand({
      userId: addUsernameRequest.userId,
      username: {
        name: addUsernameRequest.name,
        type: addUsernameRequest.type,
      },
    });
    UsernameEntity.create = jest.fn().mockReturnValue({
      id: 'john.doe@email.com',
      getProps: jest.fn().mockImplementation(() => ({
        userId: '165192d4-398a-4469-a16b-98c02cc6f531',
        name: 'john.doe@email.com',
        type: Type.EMAIL,
      })),
    });
    it('should add a new username', async () => {
      const result: AggregateID = await addUsernameService.execute(
        addUsernameCommand,
      );
      expect(result).toBe('john.doe@email.com');
    });
    it('should throw a dedicated exception if username already exists for this type', async () => {
      await expect(
        addUsernameService.execute(addUsernameCommand),
      ).rejects.toBeInstanceOf(UsernameAlreadyExistsException);
    });
    it('should throw a dedicated exception if username already exists for this name', async () => {
      await expect(
        addUsernameService.execute(addUsernameCommand),
      ).rejects.toBeInstanceOf(UsernameAlreadyExistsException);
    });
    it('should throw an error if something bad happens', async () => {
      await expect(
        addUsernameService.execute(addUsernameCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });
});
