import { IdResponse } from '@mobicoop/ddd-library';
import { RpcExceptionCode } from '@mobicoop/ddd-library';
import { UsernameAlreadyExistsException } from '@modules/authentication/core/domain/authentication.errors';
import { Type } from '@modules/authentication/core/domain/username.types';
import { AddUsernameGrpcController } from '@modules/authentication/interface/grpc-controllers/add-username.grpc.controller';
import { AddUsernameRequestDto } from '@modules/authentication/interface/grpc-controllers/dtos/add-username.request.dto';
import { CommandBus } from '@nestjs/cqrs';
import { RpcException } from '@nestjs/microservices';
import { Test, TestingModule } from '@nestjs/testing';

const addUsernameRequest: AddUsernameRequestDto = {
  userId: '78153e03-4861-4f58-a705-88526efee53b',
  name: 'john.doe@email.com',
  type: Type.EMAIL,
};

const mockCommandBus = {
  execute: jest
    .fn()
    .mockImplementationOnce(() => 'john.doe@email.com')
    .mockImplementationOnce(() => {
      throw new UsernameAlreadyExistsException();
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Add Username Grpc Controller', () => {
  let addUsernameGrpcController: AddUsernameGrpcController;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: CommandBus,
          useValue: mockCommandBus,
        },
        AddUsernameGrpcController,
      ],
    }).compile();

    addUsernameGrpcController = module.get<AddUsernameGrpcController>(
      AddUsernameGrpcController,
    );
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(addUsernameGrpcController).toBeDefined();
  });

  it('should add a new username', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    const result: IdResponse = await addUsernameGrpcController.addUsername(
      addUsernameRequest,
    );
    expect(result).toBeInstanceOf(IdResponse);
    expect(result.id).toBe('john.doe@email.com');
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if username already exists', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await addUsernameGrpcController.addUsername(addUsernameRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.ALREADY_EXISTS);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a generic RpcException', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await addUsernameGrpcController.addUsername(addUsernameRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.UNKNOWN);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });
});
