import { RpcExceptionCode } from '@mobicoop/ddd-library';
import {
  Action,
  Domain,
} from '@modules/authorization/core/domain/authorization.types';
import { DecisionResponseDto } from '@modules/authorization/interface/dtos/decision.response.dto';
import { DecideGrpcController } from '@modules/authorization/interface/grpc-controllers/decide.grpc.controller';
import { DecisionRequestDto } from '@modules/authorization/interface/grpc-controllers/dtos/decision.request.dto';
import { QueryBus } from '@nestjs/cqrs';
import { RpcException } from '@nestjs/microservices';
import { Test, TestingModule } from '@nestjs/testing';

const decisionRequest: DecisionRequestDto = {
  domain: Domain.USER,
  action: Action.READ,
  context: [
    {
      name: 'owner',
      value: '96d99d44-e0a6-458e-a656-de2a400d60a8',
    },
    {
      name: 'id',
      value: '96d99d44-e0a6-458e-a656-de2a400d60a8',
    },
  ],
};

const mockQueryBus = {
  execute: jest
    .fn()
    .mockImplementationOnce(() => true)
    .mockImplementationOnce(() => false)
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Decide Grpc Controller', () => {
  let decideGrpcController: DecideGrpcController;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: QueryBus,
          useValue: mockQueryBus,
        },
        DecideGrpcController,
      ],
    }).compile();

    decideGrpcController =
      module.get<DecideGrpcController>(DecideGrpcController);
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(decideGrpcController).toBeDefined();
  });

  it('should return a positive decision response', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    const decisionResponse: DecisionResponseDto =
      await decideGrpcController.decide(decisionRequest);
    expect(decisionResponse.allow).toBeTruthy();
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should return a negative decision response', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    const decisionResponse: DecisionResponseDto =
      await decideGrpcController.decide(decisionRequest);
    expect(decisionResponse.allow).toBeFalsy();
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if authorization fails', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    expect.assertions(3);
    try {
      await decideGrpcController.decide(decisionRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.PERMISSION_DENIED);
    }
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
  });
});
